import logging
import mimetypes
import os
from abc import ABC, abstractmethod
from io import BytesIO
from typing import Optional, Union

import streamlit as st
from minio import Minio
from minio.commonconfig import CopySource
from minio.credentials import IamAwsProvider
from minio.credentials.providers import StaticProvider
from minio.error import MinioException
from minio.helpers import ObjectWriteResult
from pydantic import BaseModel

from xt_st_common.config import StorageType, StreamlitBaseSettings

from .utils import sizeof_fmt

settings = StreamlitBaseSettings()


class FileRef(BaseModel):
    size: Optional[str]  # e.g. '3 KB'
    size_bytes: Optional[int]
    url: Optional[str] = None  # download url
    gs_uri: Optional[str] = None  # GSC Uri
    path: str  # path to current item (e.g. /folder1/someFile.txt)

    name: str
    content_type: Optional[str]

    def set_size(self, _bytes: int):
        self.size = sizeof_fmt(_bytes)
        self.size_bytes = _bytes

    def get_root(self):
        return "/".join(self.path.split("/")[:2])

    def get_prefix(self):
        return self.path.rsplit("/", 1)[0]

    def get_folder(self):
        return self.path.rsplit("/", 2)[1]

    def get_user_folders(self):
        """Get all the folders between the project root and the filename.

        Returns:
            str: User folders separated by "/"
        """
        folders = self.path.split("/")

        if len(folders) < 3:
            return ""

        return "/".join(folders[2:-1])

    def get_suffix(self):
        return self.path.split(self.get_prefix())[1]

    def get_ext(self):
        parts = self.name.rsplit(".", 1)
        return None if len(parts) <= 1 else f"{parts[-1]}"


class StorageClient(ABC):
    @abstractmethod
    def write_file(self, file_path: str, file_data: str, content_type="text/plain") -> FileRef:
        pass

    @abstractmethod
    def get_file(self, file_name):
        pass

    @abstractmethod
    def delete_file(self, file_name):
        pass

    @abstractmethod
    def list_files(self, path_prefix=""):
        pass

    @abstractmethod
    def file_exists(self, file_path: str):
        pass

    @abstractmethod
    def copy_file(
        self,
        file_path: str,
        new_file_path: str,
        delete_original: bool = False,
    ):
        pass


class MinioStorageClient(StorageClient):
    def __init__(
        self,
        bucket: str,
        endpoint="s3.amazonaws.com",
        credentials=IamAwsProvider(),
        secure=True,
    ):
        if not bucket:
            raise ValueError("bucket name must be set")

        self.__bucket = bucket
        # Check environment variables have been set
        if isinstance(credentials, IamAwsProvider):
            for var in ["AWS_WEB_IDENTITY_TOKEN_FILE", "AWS_REGION", "AWS_ROLE_ARN"]:
                if os.environ.get(var) is None or len(str(os.environ.get(var))) <= 0:
                    raise ValueError("Environment Variable {var} missing")

        # Create the Minio Client
        __client = Minio(endpoint, credentials=credentials, secure=secure)
        exists = __client.bucket_exists(bucket)
        if not exists:
            __client.make_bucket(bucket)
            logging.info(f"Minio Bucket created: {bucket}")

        self.__client = __client

    @classmethod
    def object_to_file(
        cls,
        obj: ObjectWriteResult,
        file_length: int = 0,
        content_type: str = "text/plain",
    ) -> FileRef:
        path = obj.object_name
        path_segments = path.split("/")
        name = path_segments[-1]

        return FileRef(
            url=obj.bucket_name + obj.object_name,
            size=sizeof_fmt(file_length),
            size_bytes=file_length,
            path=path,
            name=name,
            content_type=content_type,
        )

    def write_file(self, file_path: str, file_data: Union[str, bytes], content_type=None) -> Optional[FileRef]:
        if isinstance(file_data, str):
            file_bytes = BytesIO(file_data.encode("utf-8"))
            file_len = len(file_data)
            # print(f"STR: {file_len}; {file_data}")
        elif isinstance(file_data, BytesIO):
            file_bytes: BytesIO = file_data
            file_len = file_bytes.getbuffer().nbytes
            # print(f"BytesIO: {file_len}")
        else:
            file_bytes = BytesIO(file_data)
            file_len = len(file_data)

        if content_type is None:
            content_type = mimetypes.guess_type(file_path)[0]

        # set MIME_TYPE for MSCL/GeoTek files
        if content_type is None and file_path.lower().endswith((".cal", ".out", ".raw", ".sbd", ".xrf")):
            content_type = "text/plain"

        obj = self.__client.put_object(self.__bucket, file_path, file_bytes, file_len, content_type=content_type)
        return self.object_to_file(obj, file_len, content_type)

    def copy_file(
        self,
        file_path: str,
        new_file_path: str,
        content_type,
        size_bytes: int,
        delete_original: bool = False,
    ) -> Optional[FileRef]:
        obj = self.__client.copy_object(self.__bucket, new_file_path, CopySource(self.__bucket, file_path))

        new_file_ref = self.object_to_file(obj, file_length=size_bytes, content_type=content_type)

        if delete_original:
            self.delete_file(file_path)

        return new_file_ref

    def get_file(self, file_path: str):
        response_data = None
        try:
            response = self.__client.get_object(self.__bucket, file_path)
            # Read data from response.
            response_data = BytesIO(response.read())
            response_data.name = file_path
        finally:
            response.close()
            response.release_conn()

        return response_data

    def delete_file(self, file_path: str):
        self.__client.remove_object(self.__bucket, file_path)

    def list_files(self, path_prefix="", recursive=True):
        return self.__client.list_objects(self.__bucket, prefix=path_prefix, recursive=recursive)

    def file_exists(self, file_path: str):
        try:
            stats = self.__client.stat_object(self.__bucket, file_path)
            return stats is not None
        except MinioException as err:
            if hasattr(err, "code") and err.code == "NoSuchKey":
                return False
            logging.error(str(err))
            raise err


@st.cache_resource()
def storage_client():
    if settings.STORAGE_TYPE not in [StorageType.MINIO, StorageType.MINIO_AWS]:
        raise ValueError("STORAGE_TYPE may not be blank when using the storage_client")
    if settings.BUCKET_NAME is None:
        raise ValueError("BUCKET_NAME may not be blank!")
    if settings.STORAGE_TYPE == StorageType.MINIO_AWS:
        return MinioStorageClient(settings.BUCKET_NAME)  # type: ignore
    if settings.STORAGE_TYPE == StorageType.MINIO:
        if settings.MINIO_ENDPOINT is None:
            raise ValueError("MINIO_ENDPOINT may not be blank!")
        if settings.MINIO_SECRET_KEY is None:
            raise ValueError("MINIO_SECRET_KEY may not be blank!")
        if settings.MINIO_ACCESS_KEY is None:
            raise ValueError("MINIO_ACCESS_KEY may not be blank!")
        return MinioStorageClient(
            settings.BUCKET_NAME,
            endpoint=settings.MINIO_ENDPOINT,
            credentials=StaticProvider(settings.MINIO_ACCESS_KEY, settings.MINIO_SECRET_KEY.get_secret_value()),
            secure=settings.MINIO_HTTPS,
        )
    raise NotImplementedError(f"Storage Type {settings.STORAGE_TYPE} not supported")



import logging
from enum import Enum
from pathlib import Path
from typing import Annotated, Any, Dict, List, Literal, Optional, Union

import natsort
from bson import ObjectId
from pydantic import AfterValidator, BaseModel, ConfigDict, Field, PlainSerializer, WithJsonSchema

from xt_st_common.config import StreamlitBaseSettings
from xt_st_common.storage import FileRef, storage_client

settings = StreamlitBaseSettings()


def validate_object_id(v: Any) -> ObjectId:
    if isinstance(v, ObjectId):
        return v
    if ObjectId.is_valid(v):
        return ObjectId(v)
    raise ValueError("Invalid ObjectId")


PyObjectId = Annotated[
    Union[str, ObjectId],
    AfterValidator(validate_object_id),
    PlainSerializer(lambda x: str(x), return_type=str),
    WithJsonSchema({"type": "string"}, mode="serialization"),
]


class ProjectState(str, Enum):
    PROJECT_DELETE_CONFIRM = "proj_delete_confirm"
    PROJECT_TO_DELETE = "proj_delete"
    PROJECT_SUCCESS_MESSAGE = "proj_success_message"
    PROJECT_WARNING_MESSAGE = "proj_warning_message"

    FOLDER_TO_DELETE = "upload_delete_folder"
    FOLDER_ADDED = "upload_folder_added"

    UPLOAD_SUCCESS_MESSAGE = "upload_success_message"
    UPLOAD_WARNING_MESSAGE = "upload_warning_message"
    UPLOAD_DELETE_CONFIRM = "upload_delete_confirm"

    FILE_DELETE_CONFIRM = "file_delete_confirm"
    FILE_TO_DELETE = "file_to_delete"
    FILE_SUCCESS_MESSAGE = "file_success_message"
    FILE_WARNING_MESSAGE = "file_warning_message"
    FILE_MESSAGE = "file_message"

    # FILE_MANAGER_UPLOAD = "file_manager_upload_file"
    FILE_MANAGER_REPLACE_FILE = "file_manager_replace_file"

    def __str__(self):
        return str(self.value)


class Project(BaseModel):
    id: Optional[PyObjectId] = Field(None, alias="_id")
    name: str
    description: str = ""
    public: bool = False
    application: str = settings.APP_NAME
    files: List[FileRef] = []
    folders: List[str] = []
    owner: str = ""
    users: List[str] = []  # Users with access to this project

    model_config = ConfigDict(arbitrary_types_allowed=True, str_strip_whitespace=True)

    def get_users_string(self):
        return ",".join(self.users)

    def get_folders_map(self):
        fol_dict = {"": "Project Root"}
        for folder in self.folders:
            parts = folder.strip("/").split("/")
            fol_dict[folder] = " - ".join(parts)
        return fol_dict

    def get_folder_path(self, folder):
        return f"{self.id!s}/{self.name}/{folder}".strip("/")

    def get_files_in_folder(
        self,
        folder: str,
        include_subfolders: bool = True,
        extensions: Optional[List[str]] = None,
        null_option: Optional[str] = None,
        sort: Literal["none", "file-asc", "file-desc", "folder-asc", "folder-desc"] = "none",
    ):
        """
        _summary_.

        Parameters
        ----------
        folder : str
            The folder to search, set to '' to search the root folder.
        include_subfolders : bool, optional
            If true will find all files in subfolders, if false will only search the specified folder, by default True
        extensions : List[str], optional
            Extensions to match on, MUST BE LOWERCASE without a '.' use None to not filter on extensions,
            by default None
        null_option : _type_, optional
            Will add a default 'None' item to the list if None no default item will be added, by default None
        sort : str, optional
            Will enable sorting the file list by only filename or by parent folder in ascending or descending order
        Returns
        -------
        _type_
            _description_
        """

        _files: Dict[str, Union[None, FileRef]] = {} if null_option is None else {null_option: None}

        path = self.get_folder_path(folder)
        for file in self.files:
            if path in file.path:
                file_path = file.path.removeprefix(path).strip("/")
                if (include_subfolders or file_path.find("/") <= 0) and (
                    extensions is None or str(Path(file.name).suffix).strip(".").lower() in extensions
                ):
                    _files[file_path] = file

        if sort == "file-asc":
            _files = natsort.natsorted(_files, reverse=False, key=lambda x: _files[x].name)
        elif sort == "file-desc":
            _files = natsort.natsorted(_files, reverse=True, key=lambda x: _files[x].name)
        elif sort == "folder-asc":
            _files = natsort.natsorted(_files, reverse=False, key=lambda x: _files[x].get_user_folders())
        elif sort == "folder-desc":
            _files = natsort.natsorted(_files, reverse=True, key=lambda x: _files[x].get_user_folders())

        return _files

    def get_subfolders(self, folder: str, include_all: bool = False) -> list:
        """
        Get subfolders of the provided search folder.

        Args:
            folder (str): The name of the folder to search for subfolders
            include_all(bool): If False, return only the immediate children of the search folder.
                If True, return *all* sub folders ie. the children of children. Default False.

        Returns:
            list[str]: A list of sub folder strings
        """

        if include_all:
            # when split on the folder name, a subfolder will have "" as the first split entry.
            return [sub for sub in self.folders if sub.split(f"{folder}/")[0] == ""]

        return [
            # do as above, but don't accept subfolders that contain "/" (and are therefore children)
            sub
            for sub in self.folders
            if (sub.split(f"{folder}/")[0] == "") and "/" not in sub.split(f"{folder}/")[1]
        ]

    def populate_node(self, _dict, part, parts):
        if part not in _dict:
            _dict[part] = {}

        if parts:
            self.populate_node(_dict[part], parts[0], parts[1:])

    def populate_children(self, _dict, key, children, path, show_checkbox=True):
        if child_dict := _dict[key]:
            for child_key in child_dict:
                new_children = None
                new_path = f"{path}/{child_key}"
                if child_dict[child_key]:
                    new_children = []
                    self.populate_children(child_dict, child_key, new_children, new_path, show_checkbox)
                children.append(
                    {
                        "label": child_key,
                        "value": f"{path}/{child_key}",
                        "children": new_children,
                        "showCheckbox": show_checkbox,
                    }
                )

    def get_file_tree(self, show_checkbox=True):
        main_dict = {}
        full_list = (
            [file.path.removeprefix(self.get_folder_path("")).strip("/") for file in self.files] if self.files else []
        )
        if self.folders:
            full_list.extend(self.folders)
        if not full_list:
            return []
        for folder in full_list:
            parts = folder.split("/")
            self.populate_node(main_dict, parts[0], parts[1:])

        nodes = []
        for key in main_dict:
            children = None
            if main_dict[key]:
                children = []
                self.populate_children(main_dict, key, children, key, show_checkbox)
            nodes.append({"label": key, "value": key, "children": children, "showCheckbox": show_checkbox})

        return nodes

    def add_replace_file(self, data_file, folder: str, filename: str, content_type=None) -> FileRef:
        path = f"{self.get_folder_path(folder)}/{filename}".strip("/")
        return self.add_replace_file_by_path(data_file, path, content_type=content_type)

    def add_replace_file_by_path(self, data_file, path: str, content_type=None) -> FileRef:
        try:
            file_ref = storage_client().write_file(path, data_file, content_type)
            replaced = False
            for idx, file in enumerate(self.files):
                if file.path == path:
                    self.files[idx] = file_ref
                    replaced = True
                    break
            if not replaced:
                self.files.append(file_ref)
            return file_ref
        except Exception as err:
            logging.error(err)
            raise err

    def move_file_to_path(self, file: FileRef, new_path: str) -> FileRef:
        """
        Move the file referenced by file_ref to the new_path.

        Args:
            file_ref (FileRef): A file reference object for the existing file
            new_path (str): The new path where the file to be moved to

        Returns:
            FileRef: The file reference for the newly moved file.
        """
        new_file_ref = storage_client().copy_file(
            file.path,
            new_path,
            file.content_type,
            file.size_bytes,
            delete_original=True,
        )
        index = self.files.index(file)
        self.files[index] = new_file_ref
        return new_file_ref

    def copy_file_to_path(self, file_ref: FileRef, new_path: str) -> FileRef:
        """
        Copy the file referenced by file_ref to the new_path.

        Args:
            file_ref (FileRef): A file reference object for the existing file
            new_path (str): The new path where the file to be copied to

        Returns:
            FileRef: The file reference for the newly copied file.
        """
        new_file_ref = storage_client().copy_file(
            file_ref.path,
            new_path,
            file_ref.content_type,
            file_ref.size_bytes,
            delete_original=False,
        )
        replaced = False
        for idx, file in enumerate(self.files):
            if file.path == new_path:
                self.files[idx] = new_file_ref
                replaced = True
                break
        if not replaced:
            self.files.append(new_file_ref)
        return new_file_ref

    def delete_file(self, file: FileRef):
        storage_client().delete_file(file.path)
        try:
            self.files.remove(file)
        except ValueError:
            logging.warning(f"File {file.path} didn't exist and couldn't be deleted.")

    def add_folders(self, folders_string: str):
        folders = set(self.folders)
        new_folders = folders_string.split(",")
        count = 0
        for folder in new_folders:
            folder = folder.strip().replace(" ", "_").strip("/")
            folder_parts = folder.split("/")
            # add each part
            for idx, part in enumerate(folder_parts):
                part_string = part if idx == 0 else f"{'/'.join(folder_parts[:idx + 1])}"

                # only increment count if the part_string doesn't already exist in the folder list
                if part_string not in folders:
                    folders.add(part_string)
                    count += 1

        try:
            self.folders = list(folders)
        except AttributeError:
            logging.warning("Weird odmatic error, attempting to ignore")

        return count

    def delete_folder(self, folder: str):
        """
        Recursively delete the files and sub folders of the provided folder name.

        Args:
            folder (str): The name of the folder to be recursively deleted
        """
        try:
            sub_folders = self.get_subfolders(folder)

            # recursively delete any sub folders first
            for sub_folder in sub_folders:
                # this "if" appears to be redundant but the sub-folder may already have been
                # deleted in another recursion branch
                if sub_folder in self.folders:
                    self.delete_folder(sub_folder)

            folder_path = self.get_folder_path(folder)
            self.folders.remove(folder)
            for file in [x for x in self.files if folder_path in x.path]:
                self.delete_file(file)
        except ValueError:
            logging.warning(f"Folder {folder} didn't exist and couldn't be deleted.")

    def rename_folder(self, existing_name: str, new_name: str):
        """
        Rename the folder with existing name with new_name and move all of its files.

        Args:
            existing_name (str): The name of the existing folder
            new_name (str): The new path where the file to be moved to

        Returns:
            FileRef: The file reference for the newly moved file.
        """

        sub_folders = self.get_subfolders(existing_name)

        # recursively move any sub folders first
        for sub_folder in sub_folders:
            # this "if" appears to be redundant but the sub-folder may already have been
            # moved in another recursion branch
            if sub_folder in self.folders:
                self.rename_folder(sub_folder, sub_folder.replace(existing_name, new_name))

        files = self.get_files_in_folder(existing_name, include_subfolders=False)
        for file_ref in files.values():
            self.move_file_to_path(file_ref, "/".join([file_ref.get_root(), new_name, file_ref.name]))
        self.folders[self.folders.index(existing_name)] = new_name

    def rename_project(self, existing_name: str, new_name: str):
        """
        Rename the current project and update all of its file paths.

        Args:
            existing_name (str): The existing name of the project
            new_name (str): The name to rename the project as
        """
        for file_ref in self.files:
            (
                self.move_file_to_path(
                    file_ref,
                    "/".join(
                        [
                            file_ref.get_root().replace(existing_name, new_name),
                            file_ref.get_user_folders(),
                            file_ref.name,
                        ]
                    ).replace("//", "/"),
                ),
            )

        self.name = new_name



import json
from io import BytesIO
from pathlib import Path
from random import randint
from typing import List, Literal, Optional, Tuple, Union  # , Optional

import pandas as pd
import streamlit as st

# from mpl_toolbox_ui.common.config import Settings
from xt_st_common.database import (
    Project,
    get_project_cache,
    save_project,
)
from xt_st_common.project_actions import (
    action_delete,
    action_delete_file,
    action_delete_folder,
    add_folders,
    copy_file_to_project,
    get_df_preview,
    get_proj_options,
    get_selected_folder,  # noqa: F401
    get_selected_project,
    get_string_preview,
    has_project_write_access,
    move_file,
    move_folder,
    parse_csv_data,
    set_selected_project,
    state_reset,
    submit_delete_project,
    submit_edit_projects,
)
from xt_st_common.project_models import ProjectState
from xt_st_common.session import get_user_email
from xt_st_common.storage import storage_client
from xt_st_common.utils import (
    get_encoding_and_dialect,
    get_state,
    set_state,
)

# from bson.objectid import ObjectId

FOLDER_SHARED_KEY = "project_folder_select"
PROJECT_SHARED_KEY = "project_select"

# mapping for language options for streamlit code formatter. For other available options
# see: https://github.com/react-syntax-highlighter/react-syntax-highlighter/blob/master/AVAILABLE_LANGUAGES_PRISM.MD
CODE_FORMAT_MAPPING = {
    "json": "json",
    "yaml": "yaml",
    "yml": "yaml",
    "toml": "toml",
    "md": "markdown",
}

REPLACE_FILE_HELP_TXT = (
    "The new file can have a different name but must have the same extension."
    "Warning: Uploading a new file that is significantly different to the original "
    "can have catastrophic results."
)


def _state_name(project_id: str, folder: str) -> str:
    return f"{project_id}-{folder}_fs"


def project_filters():
    with st.expander("Project Filters"):
        st.checkbox("Public projects", value=False, key="include_public_projects")
        st.checkbox("Other Apps", value=True, key="include_other_apps")
        st.checkbox("Shared projects", value=True, key="include_shared_projects")


def project_selector(
    select_box_label="Select Project",
    select_box_help=None,
    null_option="-- Select Project --",
    st_context=st.sidebar,
    on_select_change=None,
    enable_filters=True,
    set_selected=True,
    selected_in_options=True,
    prefix="",
    render_layout: Literal["vertical", "horizontal", "compact"] = "vertical",
) -> Tuple[Union["Project", None], List["Project"]]:
    """UI to select and create projects."""
    selected_project = None

    if render_layout == "horizontal":
        layout_container1, layout_container2 = st_context.columns(2)
        layout_container2.markdown("##")
    else:
        layout_container1 = st_context.container()
        layout_container2 = st_context.container()

    include_public = bool(get_state("include_public_projects", False))
    include_other_apps = bool(get_state("include_other_apps", False))
    include_shared_projects = bool(get_state("include_shared_projects", False))

    if enable_filters:
        with layout_container2:
            project_filters()
    options, projects, sel_idx = get_proj_options(
        include_public,
        include_other_apps,
        include_shared_projects,
        selected_in_options=selected_in_options,
        cache_id=get_project_cache(),
    )
    proj_options = {}

    # accomodate the null option if one is provided
    if null_option is not None:
        proj_options = {-1: null_option}
        sel_idx = sel_idx + 1

    if options is not None:
        proj_options = {**proj_options, **options}

    if len(proj_options) > 0:
        proj_idx = layout_container1.selectbox(
            select_box_label,
            help=select_box_help,
            key=f"{prefix}{PROJECT_SHARED_KEY}",
            index=sel_idx,
            on_change=on_select_change,
            options=proj_options.keys(),
            format_func=lambda x: proj_options[x],
        )

        selected_project = None
        if proj_idx is not None:
            selected_project = projects[proj_idx] if proj_idx != -1 else None
            if set_selected:
                set_selected_project(selected_project)

    else:
        st_context.warning("No projects were found")

    return selected_project, projects


def load_csv(
    data_file,
    st_context=st,
):
    """Takes a csv file and loads it into session_state."""

    try:
        encoding, dialect = get_encoding_and_dialect(data_file)
        raw_df = pd.read_csv(
            data_file,
            header=None,
            skip_blank_lines=True,
            engine="python",
            sep=None,
            encoding=encoding,
        )
    except Exception as err:
        raise ValueError("Could not parse txt/csv file.") from err

    c1, c2 = st.columns([1, 3])
    c1.header("CSV Data Import")
    c2.subheader("Import Preview (15 Rows)")
    if len(raw_df.columns) < 5:
        c2.info(
            "If preview has not loaded rows/columns correctly it may mean the wrong separator has been "
            + "detected. If that is the case than please check your file and remove unnecessary "
            + "header information."
        )
    c2.write(raw_df.head(15))

    with c1.form("config_df"):
        row_options: List[Union[str, int]] = list(range(16))
        row_options_wnone = row_options.copy()
        row_options_wnone.insert(0, "None")
        header_row = st_context.selectbox(label="Column Names Row", options=row_options)
        units_row = st_context.selectbox(label="Units Row", options=row_options_wnone)
        skip_rows = st_context.multiselect(label="Skip Rows", options=list(range(9)))

        if st.form_submit_button("Save Data"):
            return parse_csv_data(raw_df, header_row, units_row, skip_rows)
    return None, None


def _update_key(prefix: str = "", replace: bool = False):
    """Hack to clear file upload after save by updating the key."""
    value = prefix + str(randint(1000, 100000000))
    set_state(f"{prefix}file_manager_key{'_replace' if replace else ''}", value)
    return str(value)


def _get_key(prefix: str = "", replace: bool = False):
    """Hack to clear file upload after save by updating the key."""
    key = get_state(f"{prefix}file_manager_key{'_replace' if replace else ''}", None)
    return _update_key(prefix, replace) if key is None else key


def file_manager(
    project: Project,
    types: List[str],
    st_context=st.sidebar,
    upload_label: str = "Upload file(s) to selected folder",
    help_text: Optional[str] = None,
    allow_upload=True,
    allow_multiple_uploads=False,
    allow_delete=True,
    allow_replace=True,
    allow_file_rename=False,
    allow_file_move=False,
    allow_file_copy_to_project=False,
    allow_folder_add=False,
    allow_folder_rename=False,
    key_prefix: str = "",
    allow_file_select=True,
    expand_file_actions=True,
    expand_folder_actions=True,
    folder_select_text="Select Borehole/Run",
    auto_parse_csv: Optional[bool] = None,
    render_layout: Literal["vertical", "horizontal", "compact"] = "vertical",
):
    file_success_message = get_state(ProjectState.FILE_SUCCESS_MESSAGE)
    file_warning_message = get_state(ProjectState.FILE_WARNING_MESSAGE)

    folder_success_message = get_state(ProjectState.UPLOAD_SUCCESS_MESSAGE)
    folder_warning_message = get_state(ProjectState.UPLOAD_WARNING_MESSAGE)

    st_context.subheader(f"Files: {project.name}")

    if file_success_message:
        st.success(file_success_message)
    if file_warning_message:
        st.warning(file_warning_message)

    if folder_success_message:
        st.success(folder_success_message)
    if folder_warning_message:
        st.warning(folder_warning_message)

    if render_layout == "horizontal":
        layout_container1, layout_container2 = st_context.columns(2)
    else:
        layout_container1 = st_context.container()
        layout_container2 = st_context.container()

    folders_dict = project.get_folders_map()
    if allow_folder_add:
        folder_add_sub_text_container, folder_add_sub_button_container = layout_container1.columns([3, 1])
        folder_add_sub_button_container.markdown("#")
        folder_add_sub_text_container.text_input(
            "Add sub folders (use ',' to separate multiple folders and '/' to separate levels) ",
            key="add_project_folder_name",
            help="All folders are created relative to the project root. "
            + "Create multiple folders at once by using ',' as a separator, folders can be "
            + "multiple levels deep using '/' e.g. folder1/subfolder1, folder1/subfolder22",
            placeholder="folder1/subfolder1, folder1/subfolder2",
        )
        folder_add_sub_button_container.button(
            label="Add Folders",
            help="Create new folder(s)",
            on_click=add_folders,
        )

    folder = layout_container1.selectbox(
        folder_select_text,
        key=FOLDER_SHARED_KEY,
        options=folders_dict.keys(),
        format_func=lambda x: folders_dict[x],
    )
    path = project.get_folder_path(folder) if folder is not None else None
    if path is not None and folder is not None:
        row = layout_container1.expander("Folder Actions", expanded=expand_folder_actions)

        state = _state_name(str(project.id), folder)
        if state not in st.session_state:
            st.session_state[state] = 0

        if allow_upload and has_project_write_access(project):
            if auto_parse_csv is not None:
                try_parse_csv = row.checkbox(
                    "Parse CSV/TXT as Dataset",
                    value=auto_parse_csv,
                    help=(
                        "If a CSV or TXT file is uploaded you will be given options to help "
                        + "calibrate it for use as a dataset."
                    ),
                )
            else:
                try_parse_csv = False
            uploaded_files = row.file_uploader(
                upload_label,
                key=_get_key(key_prefix),
                type=types,
                help=help_text,
                accept_multiple_files=allow_multiple_uploads,
            )

            # handle cases where allow_multiple_uploads is false
            if uploaded_files is None:
                uploaded_files = []
            elif not isinstance(uploaded_files, list):
                uploaded_files = [uploaded_files]

            upload_messages = []
            for uploaded_file in uploaded_files:
                file_ref = None
                file_ref_units = None
                if (
                    uploaded_file
                    and try_parse_csv
                    and (uploaded_file.name.lower().endswith(".csv") or uploaded_file.name.lower().endswith(".txt"))
                ):
                    frame = None
                    try:
                        frame, units = load_csv(uploaded_file, st)
                    except ValueError:
                        st.warning(
                            "Could not parse CSV/TXT as a dataset. "
                            + "This may mean the file requires special parsing (such as a PWAVE file)"
                        )
                        try_parse_csv = not st.button("Upload anyway")
                        units = None

                    if frame is not None:
                        data_name = f"{Path(uploaded_file.name).stem}.feather"
                        with BytesIO() as buffer:
                            frame.to_feather(buffer)
                            buffer.seek(0)
                            file_ref = project.add_replace_file(
                                buffer,
                                folder=folder,
                                filename=data_name,
                            )
                        units_name = ""
                        if units:
                            units_name = f"{Path(uploaded_file.name).stem}_units.json"
                            units_string = json.dumps(units)
                            file_ref_units = project.add_replace_file(
                                units_string,
                                folder=folder,
                                filename=units_name,
                                content_type="application/json",
                            )

                elif uploaded_file:
                    file_ref = project.add_replace_file(
                        uploaded_file.getvalue(),
                        folder=folder,
                        filename=uploaded_file.name,
                    )

                if uploaded_file and file_ref:
                    uploaded_file.close()
                    upload_messages.append(
                        f"File: **'{file_ref.name}'** {' and ' + file_ref_units.name if file_ref_units else ''} "
                        + f" uploaded successfully to folder **'{file_ref.get_folder()}'**",
                    )

            if len(upload_messages) > 0:
                save_project(project)
                _update_key(key_prefix)
                set_state(
                    ProjectState.FILE_SUCCESS_MESSAGE,
                    "".join([f"> 1. {msg} \n" for msg in upload_messages]),
                )
                st.rerun()

        if render_layout == "compact":
            folder_add_sub_text_container = row.container()
            folder_add_sub_button_container = row.container()
            folder_delete_container = row.container()
            if allow_folder_rename:
                folder_rename_text_container = row.container()
                folder_rename_button_container = row.container()
        else:
            folder_delete_container = row.container()
            if allow_folder_rename:
                (
                    folder_rename_text_container,
                    folder_rename_button_container,
                    _,
                ) = row.columns([3, 1, 1])
                folder_rename_button_container.markdown("#")

        if allow_folder_rename and folder != "":
            folder_rename_string = folder_rename_text_container.text_input(
                "Rename folder as:",
                key="rename_project_folder_name",
                help=(
                    "Selected folder will be renamed to this value when 'Rename Selected' is clicked. "
                    + "New name will be applied to the 'leaf' folder and can not contain slashes"
                ),
                placeholder="New folder name",
            )
            if len(folder_rename_string) > 0:
                folder_rename_button_container.button(
                    label="Rename Selected",
                    help="Rename Folder",
                    on_click=move_folder,
                    args=(
                        folder,
                        folder_rename_string,
                    ),
                )

        if (
            allow_delete
            and folder
            and folder_delete_container.button(
                "Delete Selected Folder",
                key=f"{key_prefix}folder_delete_btn",
            )
        ):
            error = False
            if folder == "/":
                st.toast("Can't delete the root folder of a project", icon="⚠️")
                error = True
            elif not has_project_write_access(project):
                st.toast(f"You don't have write access to project: {project.name}", icon="⚠️")
                error = True
            if not error:
                st.warning(
                    "Deleting will remove all files that are part of this folder: "
                    + f"**'{folder}'**. Are you sure you want to continue?"
                )
                st.button("I'm Sure", on_click=action_delete_folder, args=(folder,))
        if allow_file_select:
            files = project.get_files_in_folder(folder, include_subfolders=False)
            if files is not None and len(files) > 0:
                selected_key = layout_container2.selectbox(
                    "Select File",
                    options=files.keys(),
                    key=f"{key_prefix}file_manager_file_select",
                )
                selected_file = files[selected_key] if selected_key in files else None
                if selected_file is not None and selected_key is not None:
                    row = layout_container2.expander("File Actions", expanded=expand_file_actions)
                    if len(selected_key) > 30:
                        row.caption(selected_key)

                    if render_layout == "compact":
                        file_preview_container = row.container()
                        file_prepare_container, file_download_container = row.columns([5, 3])
                        file_download_container = row.container()
                        file_rename_button_container = row.container()
                        file_delete_container = row.container()
                        if allow_file_copy_to_project:
                            file_copy_to_project_select_container = row.container()
                            file_copy_to_project_button_container = row.container()
                        if allow_file_move:
                            file_move_select_container = row.container()
                            file_move_button_container = row.container()

                    else:
                        (
                            file_delete_container,
                            file_prepare_container,
                            file_download_container,
                            file_preview_container,
                        ) = row.columns([1, 1, 1, 1])
                        (file_rename_text_container, file_rename_button_container, _) = row.columns([2, 1, 1])
                        file_rename_button_container.markdown("#")

                        if allow_file_copy_to_project:
                            (
                                file_copy_to_project_select_container,
                                file_copy_to_project_button_container,
                                _,
                            ) = row.columns([2, 1, 1])
                            file_copy_to_project_button_container.markdown("#")

                        if allow_file_move:
                            (
                                file_move_select_container,
                                file_move_button_container,
                                _,
                            ) = row.columns([2, 1, 1])
                            file_move_button_container.markdown("#")

                    # options = []
                    if allow_delete:
                        file_delete_container.button(
                            "Delete Selected",
                            key=f"{key_prefix}file_delete_btn",
                            on_click=action_delete_file,
                            args=(selected_file,),
                        )
                    if selected_key.lower().endswith((".zip", ".csv", ".geojson", ".gpkg", ".feather", ".xlsx")):
                        preview_frame = file_preview_container.button(
                            "Preview Frame",
                            key=f"{key_prefix}file_manager_preview_frame",
                        )
                        if preview_frame:
                            with st.expander(f"**Frame Viewer:** {selected_file.name}", expanded=True):
                                st.dataframe(get_df_preview(selected_file.path, selected_file.get_ext()))
                    if selected_key.lower().endswith((".json", ".yml", ".yaml", ".toml", ".md", ".txt")) and (
                        preview_frame := file_preview_container.button(
                            "Preview File",
                            key=f"{key_prefix}file_manager_preview_file",
                        )
                    ):
                        with st.expander(f"**File Viewer:** {selected_file.name}", expanded=True):
                            code_format = CODE_FORMAT_MAPPING.get(selected_key.lower().split(".")[-1])
                            if code_format is None:
                                st.write(get_string_preview(selected_file))
                            else:
                                st.code(
                                    get_string_preview(selected_file),
                                    language=code_format,
                                )
                    if file_prepare_container.checkbox(
                        "Prepare Download",
                        key=f"{key_prefix}file_manager_download_chbx",
                    ):
                        file_data = storage_client().get_file(selected_file.path)

                        file_download_container.download_button(
                            "Download",
                            file_data,
                            selected_file.name,
                            key=f"{key_prefix}file_manager_download_button",
                        )
                    if allow_replace and (
                        uploaded_replace_file := row.file_uploader(
                            "Replace the selected file",
                            key=_get_key(key_prefix, True),
                            type=selected_file.get_ext(),
                            help=REPLACE_FILE_HELP_TXT,
                            accept_multiple_files=False,
                        )
                    ):
                        project.add_replace_file_by_path(uploaded_replace_file, selected_file.path)
                        save_project(project)
                        uploaded_replace_file.close()
                        _update_key(key_prefix, True)
                        set_state(
                            ProjectState.FILE_SUCCESS_MESSAGE,
                            f"Contents of file: **'{selected_file.path}'** was succesfully replaced",
                        )
                        st.rerun()
                    if allow_file_rename:
                        rename_string = file_rename_text_container.text_input(
                            "Rename file as:",
                            key="rename_project_file_name",
                            help="Selected file will be renamed to this value when 'Rename Selected' is clicked",
                            placeholder="New file name",
                        )
                        if len(rename_string) > 0:
                            file_rename_button_container.button(
                                label="Rename Selected",
                                help="Rename File",
                                on_click=move_file,
                                args=(
                                    selected_file,
                                    f"{selected_file.get_prefix()}/{rename_string}",
                                ),
                            )
                    if allow_file_copy_to_project:
                        copy_to_selected_project, projects = project_selector(
                            select_box_label="Copy file to project:",
                            select_box_help="Selected file will be copied to this project "
                            + "when 'Copy to project' button is clicked",
                            st_context=file_copy_to_project_select_container,  # type: ignore
                            null_option=None,
                            enable_filters=False,
                            set_selected=False,
                            selected_in_options=False,
                            prefix="copy_to_project",
                        )

                        file_copy_to_project_button_container.button(
                            label="Copy to project",
                            help="Copy the selected file to the selected project. "
                            + "Folder structure of the selected file will be replicated in the new project",
                            on_click=copy_file_to_project,
                            args=(selected_file, copy_to_selected_project),
                        )

                    if allow_file_move:
                        file_move_to_folder = file_move_select_container.selectbox(
                            "Move to folder:",
                            key="move_folder_select",
                            options=folders_dict.keys(),
                            format_func=lambda x: folders_dict[x],
                        )

                        file_move_button_container.button(
                            label="Move to Folder",
                            help="Move the selected file to the selected project folder",
                            on_click=move_file,
                            args=(
                                selected_file,
                                "/".join(
                                    [
                                        selected_file.get_root(),
                                        file_move_to_folder,
                                        selected_file.name,
                                    ]
                                ).replace(
                                    "//", "/"
                                ),  # replace double slashes that occur when file_move_to_folder is project-root
                            ),
                        )

            else:
                layout_container2.markdown("##")
                layout_container2.info("No files in selected folder")

        state_reset()
    return path, folder


@st.cache_data(ttl=300)
def get_file_cached(file_path: str):
    return storage_client().get_file(file_path)


def file_selector(
    project: Project,
    folder: Optional[str] = "",
    select_folder_label: str = "Select Folder",
    select_file_label: str = "Select File",
    extensions: Optional[List[str]] = None,
    state_key: Optional[str] = None,
    prefix: str = "",
    no_file_warning: str = "No files were found with the correct extension.",
    null_option: Optional[str] = None,
):
    """
    Widget to Select and load a file from a selected project.

    Parameters
    ----------
    project : Project
        _description_
    folder : Optional[str], optional
        _description_, by default "/"
    select_folder_label : str, optional
        _description_, by default "Select Folder"
    select_file_label : str, optional
        _description_, by default "Select File"
    extensions : Optional[List[str]], optional
        _description_, by default None
    state_key : Optional[str], optional
        _description_, by default None
    no_file_warning : str, optional
        _description_, by default "No files were found with the correct extension."

    Returns
    -------
    _type_
        _description_
    """
    if folder is None:
        folders_dict = project.get_folders_map()
        folder = st.selectbox(
            select_folder_label,
            key=FOLDER_SHARED_KEY,
            options=folders_dict.keys(),
            format_func=lambda x: folders_dict[x],
        )

    files = project.get_files_in_folder(folder, extensions=extensions, include_subfolders=True, null_option=null_option)
    if files is not None and len(files) > 0:
        selected_key = st.selectbox(
            select_file_label,
            options=files.keys(),
        )
        if selected_key:
            file_ref = files.get(selected_key, None)

            if file_ref is not None and st.button(
                "Load Selected File", key=f"{state_key}_btn" if state_key else f"{prefix}load_btn_{file_ref.name}"
            ):
                file = get_file_cached(file_ref.path)
                if state_key:
                    set_state(state_key, file)
                    set_state(f"{state_key}_ref", file_ref)
                st.success(f"File {file_ref.name} loaded successfully.")
                return file, file_ref
            return None, file_ref
    else:
        st.warning(no_file_warning)
    return None, None


def get_projects_data(projects):
    selected_project = get_selected_project()
    proj_data = []
    for proj in projects:
        proj_dict = proj.dict(exclude={"files", "folders"})
        # proj_dict = proj.dict(exclude={"files", "folders", "id"})

        proj_dict["folders"] = f"{len(proj.folders) if proj.folders else 0} Folders"
        proj_dict["files"] = f"{len(proj.files) if proj.files else 0} Files"
        proj_dict["users"] = ", ".join(proj.users)
        proj_dict["select"] = proj.name == (selected_project.name if selected_project is not None else None)

        proj_data.append(proj_dict)
    return proj_data


def add_new_project(number=1, name_value: Optional[str] = None):
    user_email = get_user_email()
    if name_value is None:
        name_value = f"{user_email.split('@')[0]}_#{number}"
    project = Project(name=name_value, owner=get_user_email())

    project = save_project(project)
    set_selected_project(project)

    st.toast(f"Project **'{name_value}'** has been initialised.", icon="🎉")

    st.rerun()


def get_next_project_number(projects: list[Project]):
    highest_number = len(projects) + 1
    for proj in projects:
        try:
            proj_number = proj.name.split("#")[1]
            proj_number = int(proj_number)
            if proj_number >= highest_number:
                highest_number = proj_number + 1
        except (ValueError, IndexError):
            pass
    return highest_number


def project_manager(
    st_context=st.sidebar,
    null_option=None,
    render_layout: Literal["vertical", "horizontal", "compact"] = "vertical",
    on_project_select_change=None,
    enable_filters=True,
    auto_add_project=True,
):
    delete_confirm = get_state(ProjectState.PROJECT_DELETE_CONFIRM)
    project_to_delete = get_state(ProjectState.PROJECT_TO_DELETE, None)
    success_message = get_state(ProjectState.PROJECT_SUCCESS_MESSAGE)
    warning_message = get_state(ProjectState.PROJECT_WARNING_MESSAGE)
    get_state(ProjectState.UPLOAD_DELETE_CONFIRM)
    get_state(ProjectState.FOLDER_TO_DELETE)
    get_state(ProjectState.UPLOAD_SUCCESS_MESSAGE)
    get_state(ProjectState.UPLOAD_WARNING_MESSAGE)

    st_context.subheader("Manage Projects")

    if success_message:
        st.success(success_message)
    if warning_message:
        st.warning(warning_message)
    if delete_confirm and project_to_delete:
        st.warning(delete_confirm)
        set_state(ProjectState.PROJECT_DELETE_CONFIRM, None)
        st.button("I'm Sure", on_click=action_delete, args=(project_to_delete,))

    if render_layout == "horizontal":
        add_container, save_container, filter_container = st_context.columns([3, 1, 4])
        grid_container = st_context.container()
        select_container, delete_container, _ = st_context.columns([3, 1, 2])

        # # do some spacing that is particular to this layout
        delete_container.markdown("#")
    else:
        filter_container = st_context.container()
        grid_container = st_context.container()
        select_container = st_context.container()

        add_container = st_context.container()
        save_container, delete_container = st_context.columns([1, 1])

    if enable_filters:
        with filter_container:
            project_filters()

    with select_container:
        project, projects = project_selector(
            null_option=null_option,
            select_box_label="Select Project",
            st_context=select_container,  # type: ignore
            on_select_change=on_project_select_change,
            enable_filters=False,
        )

        if auto_add_project and len(projects) < 1:
            add_new_project()

    with grid_container:
        if render_layout == "compact":
            column_order = [
                "name",
                "description",
                "public",
                "owner",
                "users",
            ]

            set_col_width = "small"
        else:
            column_order = [
                # "select",
                # "id",
                "select",
                "name",
                "description",
                "application",
                "public",
                "owner",
                "users",
                "folders",
                "files",
            ]
            set_col_width = None  # columns will be sized to fit thier contents

        edited_project_data = st.data_editor(
            pd.DataFrame(get_projects_data(projects)),
            key="project_editor_key",
            use_container_width=True,
            hide_index=True,
            # num_rows="dynamic",
            column_order=column_order,
            column_config={
                "select": st.column_config.CheckboxColumn(
                    label="Selected",
                    help="Currently selected project",
                    width=None,
                    disabled=True,
                ),
                "name": st.column_config.TextColumn(label="Project Name", help="The name of the project", width=None),
                "description": st.column_config.Column(
                    label="Description",
                    help="General project description",
                    width=set_col_width,
                ),
                "application": st.column_config.Column(
                    label="Application",
                    help="The application that utilises this project",
                    disabled=True,
                    width=None,
                ),
                "owner": st.column_config.Column(
                    label="Owner",
                    help="The owner of the project",
                    disabled=True,
                    width=set_col_width,
                ),
                "public": st.column_config.CheckboxColumn(
                    label="Public",
                    help="Whether the project is available to all users",
                    width=None,
                ),
                "files": st.column_config.Column(
                    help="The list of files contained with the project",
                    disabled=True,
                    width=None,
                ),
                "folders": st.column_config.Column(
                    help="The list of folders contained with the project",
                    disabled=True,
                ),
                "users": st.column_config.Column(
                    label="Allowed Users (emails , separated)",
                    help="The list of users with access to the project",
                    disabled=False,
                    width=set_col_width,
                ),
            },
        )

        with add_container:
            name_col, btn_col = st.columns([3, 1])
            name_value = name_col.text_input(
                "Project Name",
                placeholder="Project_Name",
                help="Name to use for a new project",
                label_visibility="collapsed",
            )
            if btn_col.button("Add New"):
                error = False
                if not name_value:
                    error = True
                    st.toast(
                        ":red[**ADD PROJECT ERROR**:] No project name value was entered for the new project.", icon="⚠️"
                    )

                for proj in projects:
                    if proj.name.lower() == name_value.lower() and proj.owner.lower() == get_user_email().lower():
                        error = True
                        st.toast(
                            f":red[**ADD PROJECT ERROR**: Project with the name **{name_value}** already exists.",
                            icon="⚠️",
                        )
                if not error:
                    new_proj_number = get_next_project_number(projects)
                    add_new_project(new_proj_number, name_value)

        with delete_container:
            st.button(
                "Delete Selected",
                on_click=submit_delete_project,
                kwargs={"project": project},
            )

        with save_container:
            st.button(
                "Save Changes",
                disabled=(st.session_state.get("project_editor_key").get("edited_rows") is None),
                on_click=submit_edit_projects,
                kwargs={"edited_project_data": edited_project_data},
            )

    return project, projects



from typing import List, Optional

import pandas as pd
import streamlit as st

# from mpl_toolbox_ui.common.config import Settings
from xt_st_common.database import (
    Project,
    delete_project,
    get_project,
    get_project_cache,
    get_projects,
    project_duplicate_exists,
    save_project,
)
from xt_st_common.project_models import ProjectState
from xt_st_common.session import get_user_email
from xt_st_common.storage import FileRef, storage_client
from xt_st_common.utils import (
    get_encoding_and_dialect,
    get_state,
    seperate_users,
    set_state,
)


def has_project_write_access(project: Project):
    return project.owner.lower() == get_user_email().lower() or get_user_email().lower() in (
        user.lower() for user in project.users
    )


def state_reset():
    for s in ProjectState:
        if s in [ProjectState.PROJECT_TO_DELETE, ProjectState.FILE_TO_DELETE]:
            set_state(s, None)
        else:
            set_state(s, "")


def get_selected_project() -> Optional[Project]:
    project = get_state("selected_project", None)
    if project is None:
        return None
    if not isinstance(project, Project):
        raise ValueError("Selected project is not the correct type")
    return project


def get_selected_folder() -> str:
    return get_state("project_folder_select", "")


def get_selected_project_or_error() -> Project:
    project = get_selected_project()
    if project is None:
        raise ValueError("No project was selected")
    return project


def set_selected_project(project: Optional[Project]):
    if project is not None and not isinstance(project, Project):
        raise ValueError("Setting selected project to a type other than project")
    st.session_state.selected_project = project


def submit_delete_project(project: Project):
    """Callback function to set state in order to enable a delete on the next run."""
    # display a warning if the user entered an existing name

    if project is None:
        st.toast("No project is selected", icon="⚠️")
    elif project.owner.lower() != get_user_email().lower():
        st.toast(f"Delete of project **{project.name}** failed: You are not the owner.", icon="⚠️")
    else:
        set_state(
            ProjectState.PROJECT_DELETE_CONFIRM,
            (
                "Deleting will remove all files that are part of project: "
                + f"**'{project.name}'**. Are you sure you want to continue?"
            ),
        )
        set_state(ProjectState.PROJECT_TO_DELETE, project)


def action_delete(project: Project):
    if project.id is not None and project.owner.lower() == get_user_email().lower():
        delete_project(project.id)
    st.toast(f"Project **'{project.name}'** deleted successfully", icon="🎉")


def action_delete_folder(folder):
    project = get_selected_project_or_error()
    project.delete_folder(folder)
    save_project(project)
    st.toast(f"Folders **'{folder}'** was deleted successfully", icon="🎉")


# def submit_delete_file(file: FileRef):
#     """Callback function to set state in order to enable a delete on the next run."""
#     # display a warning if the user entered an existing name
#     project = st.session_state.selected_project
#     if project is None:
#         set_state(ProjectState.FILE_WARNING_MESSAGE, "No project is selected")
#     elif file is None:
#         set_state(ProjectState.FILE_WARNING_MESSAGE, "No file is selected")
#     elif not has_project_write_access(project):
#         set_state(
#             ProjectState.FILE_WARNING_MESSAGE,
#             f"You don't have write access to project: {project.name}",
#         )
#         return
#     else:
#         set_state(
#             ProjectState.FILE_DELETE_CONFIRM,
#             f"Are you sure you want to delete file **'{file.name}'**?",
#         )
#         set_state(ProjectState.FILE_TO_DELETE, file)


def action_delete_file(_file: FileRef):
    project = get_selected_project_or_error()
    if not has_project_write_access(project):
        st.toast(f"You don't have write access to project: {project.name}", icon="⚠️")
        return
    project.delete_file(_file)
    save_project(project)
    st.toast(f"File {_file.name} was deleted successfully", icon="🎉")


def submit_add_project(project: Optional[Project] = None):
    """Callback function during adding a new project."""

    message_verb = "updated"
    container = st.session_state.message_box if "message_box" in st.session_state else st

    name = get_state("create_project_name")
    description = str(get_state("create_project_description"))
    users = get_state("create_project_users")
    is_public = bool(get_state("create_project_public", False))

    if not name:
        container.warning("No project name was provided")
        return

    users_list = seperate_users(users)
    if project is None:
        message_verb = "created"
        project = Project(
            name=name,
            owner=get_user_email(),
            description=description,
            users=users_list,
            public=is_public,
        )
        if project_duplicate_exists(project.name, project.owner, str(project.id)):
            st.toast(f"A Project with the name: **'{name}'** already exists", icon="⚠️")
            return
    else:
        # Get latest DB copy of the project
        project = get_project(str(project.id), st.session_state.project_cache)
        if project is None:
            st.toast(f"Cannot update project: **'{name}'** not found in database", icon="⚠️")
            return
        project.__dict__.update(
            {
                "name": name,
                "description": description,
                "users": users_list,
                "public": is_public,
            }
        )

    project = save_project(project)
    set_selected_project(project)

    st.toast(f":green[Project **'{name}'** has been {message_verb}.]", icon="🎉")
    return


def add_folders():
    project = st.session_state.selected_project
    folders_string = st.session_state.add_project_folder_name

    if not folders_string:
        st.toast(":red[Cannot add new folders:] Folder name was empty", icon="⚠️")
        return

    if not has_project_write_access(project):
        st.toast(f"You don't have write access to project: {project.name}", icon="⚠️")
        return

    count = project.add_folders(folders_string)
    save_project(project)
    st.toast(f"{count} folders were added to **'{project.name}'**", icon="🎉")


def move_file(selected_file: FileRef, new_path: str):
    project: Project = st.session_state["selected_project"]

    if not new_path:
        st.toast("Cannot move file. New path was empty", icon="⚠️")
        return

    if selected_file.path == new_path:
        st.toast("Cannot move file. New path is the same as the existing path", icon="⚠️")
        return

    if not has_project_write_access(project):
        st.toast(f"You don't have write access to project: {project.name}", icon="⚠️")
        return

    project.move_file_to_path(selected_file, new_path)

    save_project(project)

    st.toast(f"File **'{selected_file.name}'** was succesfully moved to **'{new_path.split('/', 1)[1]}'**", icon="🎉")


def move_folder(selected_folder: str, new_folder_name: str):
    project: Project = st.session_state["selected_project"]

    if "/" in selected_folder:
        leaf_folder = selected_folder.split("/")[-1]
        folder_root = selected_folder.split(f"/{leaf_folder}")[0]
        full_new_folder_name = "/".join([folder_root, new_folder_name])
    else:
        leaf_folder = selected_folder
        folder_root = ""
        full_new_folder_name = "/".join([new_folder_name])

    if "/" in new_folder_name:
        st.toast("Cannot rename folder. New name must not contain slash (/) characters", icon="⚠️")
        return

    if not new_folder_name:
        st.toast("Cannot rename folder. New name was empty", icon="⚠️")
        return

    if leaf_folder == new_folder_name:
        st.toast("Cannot rename folder. New name is the same as the existing name", icon="⚠️")
        return

    if not has_project_write_access(project):
        st.toast(f"You don't have write access to project: {project.name}", icon="⚠️")
        return

    project.rename_folder(selected_folder, full_new_folder_name)

    save_project(project)

    st.toast(f"Folder **'{leaf_folder}'** was succesfully renamed to **'{new_folder_name}'**", icon="🎉")


def copy_file_to_project(selected_file: FileRef, new_project: Project):
    current_project: Project = st.session_state["selected_project"]

    if not has_project_write_access(new_project):
        st.toast(f"You don't have write access to project: {new_project.name}", icon="⚠️")
        return

    if current_project.name in selected_file.get_prefix():
        new_path = f"{selected_file.get_prefix().replace(current_project.name, new_project.name).replace(str(current_project.id), str(new_project.id))}{selected_file.get_suffix()}"  # noqa: E501
    else:
        st.toast("Cannot copy file. File does not exist in current project", icon="⚠️")
        return

    if selected_file.path == new_path:
        st.toast("Cannot copy file. New path is the same as the current path.", icon="⚠️")
        return

    new_file_ref = new_project.copy_file_to_path(selected_file, new_path)

    # if there are user folders included with the new path, add them to the project as well
    new_folder_message = ""
    folders_string = new_file_ref.get_user_folders()
    if folders_string != "":
        new_folder_count = new_project.add_folders(folders_string)
        if new_folder_count > 0:
            new_folder_message = f"**{new_folder_count}** folders were added."

    save_project(new_project)

    st.toast(
        f"File **'{selected_file.name}'** was succesfully copied to **'{new_project.name}'**. " + new_folder_message,
        icon="🎉",
    )


@st.cache_data(ttl=300)
def get_df_preview(path: str, ext: Optional[str], num_rows=25):
    # if filepath.suffix == ".zip":
    #     frame = get_gdf_from_file(filepath)
    #     return frame.iloc[:num_rows, :-1]
    _ext = ext.strip(".").lower() if ext else None
    if _ext == "csv":
        file = storage_client().get_file(path)
        encoding, dialect = get_encoding_and_dialect(file)
        return pd.read_csv(file, nrows=num_rows, sep=dialect.delimiter, encoding=encoding)
    if _ext == "feather":
        file = storage_client().get_file(path)
        return pd.read_feather(file)
    if _ext == "xlsx":
        file = storage_client().get_file(path)
        return pd.read_excel(file, sheet_name=0, nrows=num_rows)

    return None


def get_string_preview(fileref: FileRef):
    file = storage_client().get_file(fileref.path)
    return file.getvalue().decode("utf-8")


@st.cache_data(ttl=15)
def get_proj_options(
    include_public: bool = False,
    other_apps: bool = True,
    shared_project: bool = True,
    selected_in_options: bool = True,
    cache_id=None,
):
    selected_project = get_selected_project()
    projects = get_projects(include_public, other_apps, shared_project, get_project_cache())
    sel_idx = 0
    options = {}
    for idx, proj in enumerate(projects):
        if selected_project and proj.id == selected_project.id:
            if selected_in_options:
                sel_idx = idx
            else:
                # when selected_in_options is false, skip the rest of this iteration
                continue

        option_str = f"{proj.name} ({proj.application}){'- 📢Public' if proj.public else ''}"
        if not proj.public and proj.owner.lower() != get_user_email().lower():
            option_str += " - 🤝Shared"
        options[idx] = option_str
    return options, projects, sel_idx


def on_project_select():
    proj_idx = get_state("project_select")
    include_public = bool(get_state("include_public_projects", False))
    include_other_apps = bool(get_state("include_other_apps", False))
    options, projects, _ = get_proj_options(include_public, include_other_apps, cache_id=get_project_cache())

    if proj_idx is not None and proj_idx > -1:
        selected_project = projects[proj_idx] if proj_idx != -1 else None
        set_selected_project(selected_project)
    else:
        set_selected_project(None)


def parse_csv_data(raw_df: pd.DataFrame, header_row, units_row, skip_rows: Optional[List[int]] = None):
    units = {}
    if skip_rows is None:
        skip_rows = []

    raw_df.columns = raw_df.iloc[header_row]

    if units_row is not None and units_row != "None":
        units = raw_df.iloc[units_row].to_dict()
        if units_row != header_row and units_row not in skip_rows:
            raw_df = raw_df.drop(labels=units_row)

    if header_row not in skip_rows:
        raw_df = raw_df.drop(labels=header_row)

    raw_df = raw_df.drop(labels=skip_rows)
    raw_df = raw_df.reset_index(drop=True)

    return raw_df, units


def submit_edit_projects(edited_project_data):
    project_table_data = st.session_state.get("project_editor_key")
    for changed_id in project_table_data["edited_rows"]:
        # Get latest DB copy of the project
        db_project = get_project(
            str(edited_project_data["id"][changed_id]),
            st.session_state.project_cache,
        )
        if db_project is None:
            st.toast(
                f":orange[Cannot update project:] **'{edited_project_data['name'][changed_id]}'** "
                + " not found in database",
                icon="⚠️",
            )
        elif db_project.owner.lower() != get_user_email().lower():
            st.toast(
                f":orange[Save changes on project **{db_project.name}** failed:] You are not the owner.", icon="⚠️"
            )
        else:
            original_name = db_project.name
            db_project.__dict__.update(
                {
                    "name": edited_project_data["name"][changed_id],
                    "description": edited_project_data["description"][changed_id],
                    "users": seperate_users(edited_project_data["users"][changed_id]),
                    "public": bool(edited_project_data["public"][changed_id]),
                }
            )
            if project_duplicate_exists(db_project.name, db_project.owner, str(db_project.id)):
                st.toast(
                    f":orange[Cannot update project:] A Project with the name: **'{db_project.name}'** "
                    + "already exists",
                    icon="⚠️",
                )
            else:
                if original_name != edited_project_data["name"][changed_id]:
                    db_project.rename_project(original_name, edited_project_data["name"][changed_id])
                db_project = save_project(db_project)
                st.toast(f"Project **'{db_project.name}'** updated successfully", icon="🎉")



from enum import Enum
from importlib.machinery import PathFinder
from pathlib import Path
from typing import Optional

from pydantic import SecretStr
from pydantic_extra_types.color import Color
from pydantic_settings import BaseSettings


class StorageType(str, Enum):
    MINIO = "MINIO"
    MINIO_AWS = "MINIO_AWS"
    NONE = "NONE"


class StreamlitBaseSettings(BaseSettings):
    SIGNOUT_URL: Optional[str] = None
    CURRENT_PACKAGE: str = __name__.split(".")[0]
    BASE_PATH: str = str(Path(PathFinder().find_spec(CURRENT_PACKAGE).origin).parent)  # type: ignore
    DATA_DIR: str = str(Path.cwd() / "appdata")
    DOCS_DIR: str = str(Path.cwd() / "docs")
    """Root directory for storing app data."""

    ###############
    # Storage Vars
    ###############
    STORAGE_TYPE: StorageType = StorageType.NONE
    BUCKET_NAME: Optional[str] = None
    BUCKET_PREFIX: Optional[str] = None
    MINIO_SECRET_KEY: Optional[SecretStr] = None
    MINIO_ACCESS_KEY: Optional[str] = None
    MINIO_ENDPOINT: Optional[str] = None
    MINIO_HTTPS: bool = True

    XT_HOME_URL: str = "https://apps.exploration.tools"
    """URL of where the privacy policy button should link to."""
    XT_PRIVACY_URL: str = "https://www.csiro.au/en/About/Access-to-information/Privacy"
    """URL of where the xt logo should link to."""
    APP_NAME: str = "my_app"
    """Name of the app that be tagged in the DB."""
    APP_TAG_TEXT: str = "Dev"
    """Text that will appear next the application logo."""
    APP_TAG_BACKGROUND: Color = Color("#00a9ce")
    """Background colour of the APP_TAG_TEXT."""
    APP_GET_HELP: Optional[str] = None
    """Link to "Get Help" in streamlit hamburger menu (menu item is hidden if set to None)."""
    APP_REPORT_BUG: Optional[str] = None
    """Link to "Report a Bug" in streamlit hamburger menu (menu item is hidden if set to None)."""
    APP_ABOUT: Optional[str] = None
    """Markdown Text to show in streamlit hamburger menu (shows streamlit default if set to None)."""
    DEBUG: bool = False
    """Enable debug information in the app."""
    DEBUG_MOCK_SESSION: bool = False
    """Mock the session headers returned when deployed to AWS."""
    USE_COGNITO: bool = False
    """Use AWS Cognito for Auth."""
    ACCESS_TOKEN_KEY: str = "X-Forwarded-Access-Token"
    """Key used to retrieve the access token from the headers."""
    USER_KEY: str = "X-Forwarded-User"
    """Key used to retrieve the user from the headers."""
    EMAIL_KEY: str = "X-Forwarded-Email"
    """Key used to retrieve the email from the headers."""
    OIDC_TOKEN: str | None = None  # "X-Amzn-Oidc-Data"

    COGNITO_COOKIE: str = "AWSELBAuthSessionCookie"
    """If set cookies with this prefix will be deleted on signout."""
    COGNITO_GROUPS: str = "ap-southeast-2_lC6GUKOej_CSIRO-EASI"
    """One of these comma separated groups must be present in the user's `cognito:groups`."""
    NO_ACCESS_MSG: str = "Error: You do not have permission to use this app."
    """Error message displayed when a user doesn't have permission to use the app."""
    STREAMLIT_SERVER_BASE_URL_PATH: str = ""
    """The base path for the URL where Streamlit should be served from."""
    MONGO_CONNECTION_STRING: str = ""
    """Mongo DB connection string."""
    DATABASE_NAME: str = ""
    """Mongo DB Database Name."""
    TIMEZONE: str = "Australia/Perth"
    """Default timezone used within the app."""

    class Config:
        env_file = ".env"
        env_file_encoding = "utf-8"
        case_sensitive = True
        extra = "allow"



import streamlit as st

from xt_st_common.config import StreamlitBaseSettings

settings = StreamlitBaseSettings()


def page_header(
    page_title: str | None = None,
    page_logo: str | None = None,
    page_logo_width: int = 250,
    logout_url: str | None = None,
    header_border: bool = True,
    header_padding_top: str = "2rem",
):
    """
    Display the page header including Title, Logo, and optional Logout button.

    Args:
        page_title (str | None, optional): The title to show in the header. Defaults to None.
        page_logo (str | None, optional): A path to a page logo image. If not provided no page logo will be rendered.
            Defaults to None.
        page_logo_width (int, optional): The width of the page logo in the header. Defaults to 250.
        logout_url (str | None, optional): The logout url to be accessed when the logout button is pressed.
            If not provided the logout button will not be rendered. Defaults to None.
        header_border (bool, optional): Whether to render a single line border at the bottom of the header.
            Defaults to True.
        header_padding_top (str): adjust padding-top to move header up/down. Defaults to 2rem.
    """

    base_url = settings.STREAMLIT_SERVER_BASE_URL_PATH
    script_frame = st.empty()

    privacy_link = (
        f'<a id="privacy-link" class="btn-logout" href={settings.XT_PRIVACY_URL} target="_blank" >Privacy Policy</a>'
    )

    if logout_url is not None:
        # Load the css/icons for the button
        st.markdown(
            """
            <style scoped type="text/css">

            </style>
            """,
            unsafe_allow_html=True,
        )

        # Define the button
        logout_link = (
            f'<a id="logout-button" class="btn-logout" href={logout_url} role="button" target="_self" >🔐 Log out</a>'
        )
    else:
        logout_link = "&nbsp;"

    page_title = "" if page_title is None else f"<span class='app_title'>{page_title}</span>"

    page_logo = (
        ""
        if page_logo is None
        else f"<span class='app_logo'><img src='{page_logo}' width='{page_logo_width}px'></span>"
    )

    xt_logo = f"""<span class="xt_logo"><a href='{settings.XT_HOME_URL}'>
                <img src='{base_url}/app/static/media/XT_words_wrapped.png' width='200px'></a></span>"""

    # Display the page header including Title, Logo, and optional Logout button
    st.markdown(
        f"""
        <style scoped type="text/css">
            .main .block-container {{
                    padding-top: {header_padding_top};
            }}
            .xt_app_header {{
                {"border-bottom: 1px solid black;" if header_border else ""}
                width: 100%;
                display: flex;
                justify-content: space-between;
                align-items: center;
                flex-wrap: wrap;
                row-gap: 10px;
            }}
            .app_logo_group{{
                display: flex;
                justify-content: flex-start;
                align-items: center;
                flex-wrap: nowrap;
            }}
            .app_title{{
                font-family: arial,sans-serif;
                font-size:200%;
                font-weight:700;
                color: #575757;
                white-space: nowrap;
                line-height: 2.5rem;
                align-self: flex-end;

            }}
            .app_logo_group::after {{
                font-family: sans-serif;
                font-size: small;
                content: '{settings.APP_TAG_TEXT}';
                display: inline-block;
                text-align: center;
                color: white;
                background-color: {settings.APP_TAG_BACKGROUND};
                border-radius: 30px;
                padding: 0px 15px;
                font-weight: bold;
                align-self: start;
            }}
            .xt_logo_group{{
                display: flex;
                justify-content: space-between;
                align-items: center;
                flex-wrap: wrap;
                row-gap: 10px;
            }}
            .xt_logo{{
            }}
            .btn-logout {{
                    box-shadow:inset 0px 1px 0px 0px #ffffff;
                    background:linear-gradient(to bottom, #f9f9f9 5%, #e9e9e9 100%);
                    background-color:#f9f9f9;
                    border-radius:6px;
                    border:1px solid #dcdcdc;
                    display:inline-block;
                    cursor:pointer;
                    color:#666666 !important;
                    font-family:Arial;
                    font-size:15px;
                    padding:4px 20px !important;
                    text-decoration:none;
                    text-shadow:0px 1px 0px #ffffff;
                    margin-right: 10px;
                    flex-grow: 1;
                    text-align: center;
                }}
                .btn-logout:hover {{
                    background:linear-gradient(to bottom, #e9e9e9 5%, #f9f9f9 100%);
                    background-color:#e9e9e9;
                }}
                .btn-logout:active {{
                    top:1px;
                }}
        </style>
        """
        + "<div class='xt_app_header'>"
        + "<div class='app_logo_group'>"
        + page_logo
        + page_title
        + "</div>"
        + "<div class='xt_logo_group'>"
        + privacy_link
        + logout_link
        + xt_logo
        + "</div>"
        + "</div>",
        unsafe_allow_html=True,
    )

    # Attach a function to clear the Cognito cookies when the logout button is clicked
    if settings.USE_COGNITO and len(settings.COGNITO_COOKIE) > 0:
        from streamlit_js_eval import streamlit_js_eval

        # Attach and onClick event handler to the logout button/link
        # When the user logs out the handler will remove cookies which are prefixed with
        # `settings.COGNITO_COOKIE`.
        # Also take the opportunity to have the signout button use the current window rather
        # than opening a new one.
        logout_script = (
            """
                var a = window.parent.document.getElementById("logout-button");
                a.onclick = function() {
                    all_cookies = Object.fromEntries(document.cookie.split("; ").map(c => c.split("=")))
                    cookie_list = Object.keys(all_cookies).filter(key => key.startsWith('"""
            + settings.COGNITO_COOKIE
            + """'))
                    cookie_list.forEach(e => {
                        window.parent.document.cookie = e+'=; expires=Thu, 01 Jan 1970 00:00:01 GMT'
                    })
                    return true;
                }
                // Override the target so that we reload the current page rather than opening a new one
                a.target = ""
            """
        )
        with script_frame:
            streamlit_js_eval(js_expressions=logout_script, key="js-logout-script")



# type: ignore

import pathlib

import pytest
from pymongo import MongoClient

from xt_st_common.config import StreamlitBaseSettings

dotenv_path = pathlib.Path(__file__).parent.resolve() / ".env-test"
config = StreamlitBaseSettings(_env_file=dotenv_path)

from xt_st_common import database
from xt_st_common.project_models import Project
from xt_st_common.storage import FileRef, StorageClient
from xt_st_common.utils import sizeof_fmt

client = MongoClient(config.MONGO_CONNECTION_STRING)
collection = client.get_database(name=config.DATABASE_NAME)["project"]


class MockStorage(StorageClient):
    def write_file(self, file_path: str, file_data: str, content_type="text/plain") -> FileRef:
        return FileRef(
            url=file_path,
            size=sizeof_fmt(1000),
            size_bytes=1000,
            path=file_path,
            name=pathlib.Path(file_path).name,
            content_type=content_type,
        )

    def get_file(self, file_name):
        pass

    def delete_file(self, file_name):
        pass

    def list_files(self, path_prefix=""):
        pass

    def file_exists(self, file_path: str):
        pass

    def copy_file(
        self,
        file_path: str,
        new_file_path: str,
        file_data: str,
        content_type,
        delete_original: bool,
    ):
        return FileRef(
            url=new_file_path,
            size=sizeof_fmt(1000),
            size_bytes=1000,
            path=new_file_path,
            name=pathlib.Path(new_file_path).name,
            content_type=None,
        )


_mock_storage = MockStorage()


@pytest.fixture
def mock_collection(mocker):
    mocker.patch("xt_st_common.database.get_collection", return_value=collection)
    return mocker


@pytest.fixture
def mock_storage(mocker):
    mocker.patch("xt_st_common.project_models.storage_client", return_value=_mock_storage)
    return mocker


@pytest.fixture
def project1(mock_collection):
    OWNER = "test_user"
    NAME = "Test Project"

    project = Project(name=NAME, owner=OWNER)
    yield project
    if database.project_duplicate_exists(name=NAME, owner=OWNER):
        collection.delete_one({"name": NAME, "owner": OWNER})


@pytest.fixture
def project2(mock_collection):
    OWNER = "test_user2"
    NAME = "Test Project2"

    project = Project(name=NAME, owner=OWNER)
    yield project
    if database.project_duplicate_exists(name=NAME, owner=OWNER):
        collection.delete_one({"name": NAME, "owner": OWNER})


@pytest.fixture
def proj_with_files(mock_storage):
    OWNER = "test_user3"
    NAME = "Test Project3"
    folder = ["foo", "foo/bar"]
    project = Project(name=NAME, owner=OWNER, folders=folder)
    project.add_replace_file(data_file="I am text", folder="foo", filename="file.txt")
    project.add_replace_file(data_file="I am text", folder="foo/bar", filename="file2.json")

    return project


@pytest.fixture
def proj_with_many_folders(mock_storage):
    # a project with many folders with the same names but in different tree positions
    OWNER = "test_user4"
    NAME = "Test Project4"
    folder = [
        "foo",
        "foo/bar",
        "foo/baz",
        "foo/bar/baz",
        "foo/baz/bar",
        "foo/bar/baz/qux",
        "baz",
        "baz/foo",
        "baz/foo/bar",
    ]
    project = Project(name=NAME, owner=OWNER, folders=folder)
    project.add_replace_file(data_file="I am text", folder="/", filename="file4.txt")
    project.add_replace_file(data_file="I am text", folder="foo", filename="file1.txt")
    project.add_replace_file(data_file="I am text", folder="foo/bar", filename="file2.json")
    project.add_replace_file(data_file="I am text", folder="baz/foo/bar", filename="file3.json")

    return project


def test_project_save(project1, mock_collection):
    mock_collection.patch("xt_st_common.session.get_user_email", return_value="test_user")
    database.save_project(project1)

    assert database.project_exists(str(project1.id))


def test_get_public_success(project2: Project, mock_collection):
    project2.public = True
    mock_collection.patch("xt_st_common.session.get_user_email", return_value="test_user")
    database.save_project(project2)

    results = database.get_projects(include_public=True)

    assert results is not None, "No projects were returned, expected project2"
    assert results[0].name == project2.name, "Public project was not returned"


def test_get_public_exclude_when_false(project2: Project, mock_collection):
    project2.public = True
    mock_collection.patch("xt_st_common.session.get_user_email", return_value="test_user")
    database.save_project(project2)

    results = database.get_projects(include_public=False)

    assert not results, "A project was returned when excluding public"


def test_dont_return_private(project2: Project, mock_collection):
    project2.public = False
    mock_collection.patch("xt_st_common.session.get_user_email", return_value="test_user")
    mock_collection.patch("xt_st_common.database.get_user_email", return_value="test_user")
    database.save_project(project2)

    results = database.get_projects(include_public=True)

    assert not results, "A project was returned when project is private {}"


def test_authorised_user_can_access_project(project2: Project, mock_collection):
    project2.public = False
    project2.users = ["test_user"]
    mock_collection.patch("xt_st_common.session.get_user_email", return_value="test_user")
    mock_collection.patch("xt_st_common.database.get_user_email", return_value="test_user")
    database.save_project(project2)

    results = database.get_projects(include_public=True)

    assert results, "No project was returned"
    assert results[0].name == project2.name, "Project does not match"


def test_add_file(proj_with_files: Project):
    proj_with_files.add_replace_file(data_file="I am new text", folder="foo", filename="file3.txt")

    assert len(proj_with_files.files) == 3


def test_replace_file(proj_with_files: Project):
    proj_with_files.add_replace_file(data_file="I am new text", folder="foo", filename="file.txt")

    assert len(proj_with_files.files) == 2


def test_delete_file(proj_with_files: Project):
    proj_with_files.delete_file(proj_with_files.files[0])

    assert len(proj_with_files.files) == 1


def test_delete_deleted_file(proj_with_files: Project):
    file_ref = proj_with_files.files[0]
    proj_with_files.delete_file(file_ref)

    proj_with_files.delete_file(file_ref)

    assert len(proj_with_files.files) == 1


def test_get_folders_map(proj_with_files: Project):
    file_map = proj_with_files.get_folders_map()

    assert file_map[""] == "Project Root"
    assert file_map["foo/bar"] == "foo - bar"


def test_get_file_tree(proj_with_files: Project):
    file_tree = proj_with_files.get_file_tree()

    assert len(file_tree) == 1
    # assert len(file_tree[0]) == 3
    assert len(file_tree[0]["children"]) == 2


def test_get_files_by_extension(proj_with_files: Project):
    files = proj_with_files.get_files_in_folder(folder="", include_subfolders=True, extensions=["json"])

    assert len(files) == 1
    assert list(files.values())[0].name == "file2.json"


def test_get_files_in_folder(proj_with_many_folders: Project):
    files = proj_with_many_folders.get_files_in_folder(folder="foo/bar", include_subfolders=False, extensions=None)

    assert len(files) == 1
    assert list(files.values())[0].name == "file2.json"


def test_get_file_asc_sorted_files_in_folder(proj_with_many_folders: Project):
    files = proj_with_many_folders.get_files_in_folder(
        folder="/", include_subfolders=True, extensions=None, sort="file-asc"
    )

    assert len(files) == 4
    assert files == ["foo/file1.txt", "foo/bar/file2.json", "baz/foo/bar/file3.json", "file4.txt"]


def test_get_file_desc_sorted_files_in_folder(proj_with_many_folders: Project):
    files = proj_with_many_folders.get_files_in_folder(
        folder="/", include_subfolders=True, extensions=None, sort="file-desc"
    )

    assert len(files) == 4
    assert files == [
        "file4.txt",
        "baz/foo/bar/file3.json",
        "foo/bar/file2.json",
        "foo/file1.txt",
    ]


def test_get_folder_asc_sorted_files_in_folder(proj_with_many_folders: Project):
    files = proj_with_many_folders.get_files_in_folder(
        folder="/", include_subfolders=True, extensions=None, sort="folder-asc"
    )

    assert len(files) == 4
    assert files == [
        "file4.txt",
        "baz/foo/bar/file3.json",
        "foo/file1.txt",
        "foo/bar/file2.json",
    ]


def test_get_folder_desc_sorted_files_in_folder(proj_with_many_folders: Project):
    files = proj_with_many_folders.get_files_in_folder(
        folder="/", include_subfolders=True, extensions=None, sort="folder-desc"
    )

    assert len(files) == 4
    assert files == [
        "foo/bar/file2.json",
        "foo/file1.txt",
        "baz/foo/bar/file3.json",
        "file4.txt",
    ]


def test_add_folders(proj_with_files: Project):
    added_count = proj_with_files.add_folders("test1/test2,test3")

    assert added_count == 3
    assert len(proj_with_files.folders) == 5


def test_add_existing_folders(proj_with_files: Project):
    added_count = proj_with_files.add_folders("foo/bar")

    assert added_count == 0
    assert len(proj_with_files.folders) == 2


def test_move_file_to_path(proj_with_files: Project):
    file = proj_with_files.files[0]

    proj_with_files.move_file_to_path(file, str(file.path.replace(file.name, "renamed.txt")))

    assert proj_with_files.files[0].name == "renamed.txt"


def test_copy_file_to_path(proj_with_files: Project):
    orig_file = proj_with_files.files[0]

    proj_with_files.copy_file_to_path(orig_file, str(orig_file.path.replace(orig_file.name, "copied.txt")))

    assert len(proj_with_files.files) == 3
    assert proj_with_files.files[0].name == orig_file.name
    assert proj_with_files.files[2].name == "copied.txt"


def test_copy_existing_file_to_path(proj_with_files: Project):
    orig_file = proj_with_files.files[0]

    proj_with_files.copy_file_to_path(orig_file, str(orig_file.path.replace(orig_file.name, "copied.txt")))

    proj_with_files.copy_file_to_path(orig_file, str(orig_file.path.replace(orig_file.name, "copied.txt")))

    assert len(proj_with_files.files) == 3
    assert proj_with_files.files[0].name == orig_file.name
    assert proj_with_files.files[2].name == "copied.txt"


@pytest.mark.parametrize(
    "test_input,expected",
    [
        (
            "foo",
            2,
        ),
        (
            "foo/bar",
            1,
        ),
        (
            "foo/baz",
            1,
        ),
        (
            "baz",
            1,
        ),
        (
            "blah",
            0,
        ),
        (
            "/",
            0,
        ),
        (
            "",
            0,
        ),
    ],
)
def test_get_subfolders(test_input: str, expected: int, proj_with_many_folders: Project):
    subfolders = proj_with_many_folders.get_subfolders(test_input, include_all=False)

    assert len(subfolders) == expected


@pytest.mark.parametrize(
    "test_input,expected",
    [
        (
            "foo",
            5,
        ),
        (
            "foo/bar",
            2,
        ),
        (
            "foo/baz",
            1,
        ),
        (
            "baz",
            2,
        ),
        (
            "blah",
            0,
        ),
        (
            "/",
            0,
        ),
        (
            "",
            0,
        ),
    ],
)
def test_get_all_subfolders(test_input: str, expected: int, proj_with_many_folders: Project):
    subfolders = proj_with_many_folders.get_subfolders(test_input, include_all=True)

    assert len(subfolders) == expected


def test_delete_folders(proj_with_many_folders: Project):
    proj_with_many_folders.delete_folder("foo")

    assert len(proj_with_many_folders.folders) == 3
    assert len(proj_with_many_folders.files) == 2


@pytest.mark.parametrize(
    "from_name, to_name",
    [("foo", "foo2"), ("foo/bar", "foo/bar2"), ("baz", "baz2"), ("foo/bar/baz/qux", "foo/bar/baz/qux2")],
)
def test_rename_folder(from_name: str, to_name: str, proj_with_many_folders: Project):
    # get a count of the sub folders before the rename
    original_count = len(proj_with_many_folders.get_subfolders(from_name, include_all=True))

    proj_with_many_folders.rename_folder(from_name, to_name)

    assert original_count == len(proj_with_many_folders.get_subfolders(to_name, include_all=True))


def test_rename_project(proj_with_many_folders: Project):
    new_name = "renamed_project"

    proj_with_many_folders.rename_project(proj_with_many_folders.name, new_name)

    assert proj_with_many_folders.name == new_name
    for file_ref in proj_with_many_folders.files:
        assert new_name in file_ref.path



#!/usr/bin/env python3
"""
Release tag creation utility for xt-workbench.

This script guides developers through creating properly formatted git tags for releases.
Supports automatic semver bumping (major/minor/patch) and custom version input with safeguards.

Release types:
  - staging: Creates v<VERSION>-rc tag for staging environment (from develop or production-support branches)
  - prod: Creates v<VERSION> tag by promoting a tested -rc tag

Branch rules:
  - Staging: Must be on develop or production-support branch (allow main with warning)
  - Production: Can be on any branch (checks git, not working tree)

Changelog management:
  - Automatically updates [Unreleased] section to new version before tagging
  - Commits changelog changes before tag creation
  - Supports --force to skip changelog check

Usage:
  python scripts/release.py

Or via poe (in backend directory):
  poe release-tag

Environment:
  Requires git, Python 3.12+, working git repository with tags
"""

import re
import subprocess
import sys
from pathlib import Path
from typing import Optional, Tuple

AUTO_CONFIRM = False


class Colors:
    """ANSI color codes for terminal output."""

    RESET = "\033[0m"
    BOLD = "\033[1m"
    DIM = "\033[2m"
    RED = "\033[91m"
    GREEN = "\033[92m"
    YELLOW = "\033[93m"
    BLUE = "\033[94m"
    CYAN = "\033[96m"


def print_header(text: str) -> None:
    """Print a formatted header."""
    print(f"\n{Colors.BOLD}{Colors.BLUE}{'=' * 70}{Colors.RESET}")
    print(f"{Colors.BOLD}{Colors.BLUE}{text:^70}{Colors.RESET}")
    print(f"{Colors.BOLD}{Colors.BLUE}{'=' * 70}{Colors.RESET}\n")


def print_success(text: str) -> None:
    """Print success message."""
    print(f"{Colors.GREEN}✓ {text}{Colors.RESET}")


def print_error(text: str) -> None:
    """Print error message."""
    print(f"{Colors.RED}✗ {text}{Colors.RESET}")


def print_warning(text: str) -> None:
    """Print warning message."""
    print(f"{Colors.YELLOW}⚠ {text}{Colors.RESET}")


def print_info(text: str) -> None:
    """Print info message."""
    print(f"{Colors.CYAN}ℹ {text}{Colors.RESET}")


def run_command(cmd: str, cwd: Optional[str] = None) -> Tuple[int, str]:
    """Run a shell command and return exit code and output."""
    try:
        result = subprocess.run(cmd, shell=True, capture_output=True, text=True, cwd=cwd)
        return result.returncode, result.stdout.strip()
    except Exception as e:
        print_error(f"Failed to run command: {cmd}\nError: {e}")
        sys.exit(1)


def get_repo_root() -> str:
    """Get the root directory of the git repository."""
    returncode, output = run_command("git rev-parse --show-toplevel")
    if returncode != 0:
        print_error("Not in a git repository")
        sys.exit(1)
    return output


def get_current_branch() -> str:
    """Get the current git branch name."""
    returncode, output = run_command("git rev-parse --abbrev-ref HEAD")
    if returncode != 0:
        print_error("Failed to get current branch")
        sys.exit(1)
    return output


def get_current_version() -> Optional[str]:
    """Detect current version from git tags."""
    print_info("Detecting current version from git tags...")

    # Use semver sort across all tags (not only reachable from HEAD). This avoids
    # `git describe` picking an older reachable tag when a newer tag exists on
    # another branch (e.g., latest tag v1.6.0-rc but current branch only sees
    # v1.5.12-rc).
    returncode, output = run_command("git tag -l 'v*' --sort=-version:refname")
    if returncode == 0 and output:
        for tag in output.split("\n"):
            match = re.match(r"v?(\d+\.\d+\.\d+)(-rc)?", tag)
            if match:
                version = match.group(1)
                print_success(f"Current version: {version} (from tag: {tag})")
                return version

    print_warning("Could not detect version automatically, defaulting to 0.0.0")
    return "0.0.0"


def parse_version(version_str: str) -> Optional[Tuple[int, int, int]]:
    """Parse semantic version string into (major, minor, patch) tuple."""
    match = re.match(r"v?(\d+)\.(\d+)\.(\d+)", version_str)
    if match:
        return int(match.group(1)), int(match.group(2)), int(match.group(3))
    return None


def format_version(major: int, minor: int, patch: int) -> str:
    """Format version tuple into semantic version string."""
    return f"{major}.{minor}.{patch}"


def bump_version(version: str, bump_type: str) -> str:
    """Bump version based on type (major, minor, patch)."""
    parsed = parse_version(version)
    if not parsed:
        print_error(f"Invalid version format: {version}")
        sys.exit(1)

    major, minor, patch = parsed

    if bump_type == "major":
        major += 1
        minor = 0
        patch = 0
    elif bump_type == "minor":
        minor += 1
        patch = 0
    elif bump_type == "patch":
        patch += 1
    else:
        print_error(f"Invalid bump type: {bump_type}")
        sys.exit(1)

    return format_version(major, minor, patch)


def confirm_action(prompt: str, default: bool = True) -> bool:
    """Get user confirmation with default; auto-confirm when requested."""
    if AUTO_CONFIRM:
        return True

    default_str = "[Y/n]" if default else "[y/N]"
    response = input(f"{Colors.BOLD}{prompt} {default_str}: {Colors.RESET}").strip().lower()

    if response in ("y", "yes"):
        return True
    if response in ("n", "no"):
        return False
    return default


def choose_option(prompt: str, options: list, preselection: Optional[str] = None) -> str:
    """Let user choose from a list of options, honoring a preselection for non-interactive use."""
    if preselection:
        if preselection in options:
            return preselection
        print_error(f"Invalid selection '{preselection}'. Valid options: {', '.join(options)}")
        sys.exit(1)

    print(f"\n{Colors.BOLD}{prompt}{Colors.RESET}")
    for i, option in enumerate(options, 1):
        print(f"  {Colors.CYAN}{i}{Colors.RESET}. {option}")

    while True:
        choice = input(f"\n{Colors.BOLD}Select option [1-{len(options)}]: {Colors.RESET}").strip()
        if choice.isdigit() and 1 <= int(choice) <= len(options):
            return options[int(choice) - 1]
        print_error(f"Invalid selection. Please enter 1-{len(options)}")


def is_version_lower_than_any_tag(version: str) -> bool:
    """Check if provided version is lower than any existing tag."""
    parsed = parse_version(version)
    if not parsed:
        return False

    returncode, output = run_command("git tag -l 'v*' --sort=-version:refname")
    if returncode != 0:
        return False

    tags = output.split("\n") if output else []
    for tag in tags:
        tag_parsed = parse_version(tag)
        if tag_parsed and tag_parsed > parsed:
            return True

    return False


def check_branch_validity(release_type: str, force: bool = False) -> bool:
    """Validate current branch for release type."""
    branch = get_current_branch()

    if release_type == "staging":
        allowed_branches = ["develop", "production-support", "main"]
        if branch not in allowed_branches:
            print_error(f"Staging releases can only be created from: {', '.join(allowed_branches)}")
            print_error(f"Current branch: {branch}")
            if not force:
                return False
            print_warning("--force flag used; proceeding anyway")

        if branch == "develop":
            print_warning(
                "You are on the 'develop' branch. Consider pushing these changes to 'main' via a PR before creating a staging release."
            )
            if not confirm_action("Continue with staging release?"):
                return False

    return True


def find_changelog_file(repo_root: str) -> Optional[Path]:
    """Find changelog file (supports CHANGELOG.md, changelog.md, HISTORY.md, etc.)."""
    changelog_patterns = [
        "changelog.md",
        "CHANGELOG.md",
        "CHANGELOG.MD",
        "HISTORY.md",
        "HISTORY.md",
        "changes.md",
        "CHANGES.md",
    ]

    repo_path = Path(repo_root)
    for pattern in changelog_patterns:
        candidate = repo_path / pattern
        if candidate.exists():
            return candidate

    return None


def update_changelog(changelog_path: Path, version: str, force: bool = False) -> bool:
    """Update changelog: replace [Unreleased] with new version heading."""
    print_info(f"Checking changelog: {changelog_path.name}")

    try:
        content = changelog_path.read_text(encoding="utf-8")
    except Exception as e:
        print_error(f"Failed to read changelog: {e}")
        if not force:
            print_warning("Use --force to skip changelog update")
            return False
        print_warning("Skipping changelog update (--force flag used)")
        return True

    # Check if [Unreleased] section exists
    unreleased_pattern = r"^## \[Unreleased\]\s*$"
    if not re.search(unreleased_pattern, content, flags=re.MULTILINE):
        print_warning("No [Unreleased] section found in changelog")
        if not confirm_action("Proceed without updating changelog?"):
            return False
        return True

    # Replace first [Unreleased] heading with new version heading, re-adding [Unreleased]
    updated_content = re.sub(
        unreleased_pattern,
        f"## [Unreleased]\n\n## {version}",
        content,
        count=1,
        flags=re.MULTILINE,
    )

    if updated_content == content:
        print_warning("Changelog appears unchanged after update attempt")
        if not confirm_action("Continue with release?"):
            return False
    else:
        try:
            changelog_path.write_text(updated_content, encoding="utf-8")
            print_success(f"Updated changelog with version {version}")
        except Exception as e:
            print_error(f"Failed to write changelog: {e}")
            if not force:
                return False
            print_warning("Could not write changelog (--force flag used)")

    return True


def commit_changelog(changelog_path: Path, version: str) -> bool:
    """Commit changelog changes."""
    returncode, _ = run_command("git status --porcelain")
    if returncode != 0:
        return False

    # Check if there are staged changes
    returncode, output = run_command("git diff --cached --name-only")
    if returncode != 0 or not output:
        # Nothing staged; add the changelog
        returncode, _ = run_command(f"git add {changelog_path}")
        if returncode != 0:
            print_error("Failed to stage changelog")
            return False

    # Commit
    commit_msg = f"chore: update changelog for {version}"
    returncode, output = run_command(f'git commit -m "{commit_msg}"')
    if returncode != 0:
        # Might already be committed or no changes
        print_warning(f"Changelog commit returned code {returncode}")
        # Not fatal; continue
    else:
        print_success(f"Committed changelog: {commit_msg}")

    return True


def create_tag(tag_name: str, push: bool = False) -> bool:
    """Create git tag and optionally push to remote."""
    # Check if tag already exists
    returncode, _ = run_command(f"git rev-parse {tag_name} 2>/dev/null")
    if returncode == 0:
        print_error(f"Tag '{tag_name}' already exists!")
        return False

    # Create tag
    returncode, output = run_command(f"git tag {tag_name}")
    if returncode != 0:
        print_error(f"Failed to create tag: {output}")
        return False

    print_success(f"Created tag: {tag_name}")

    if push:
        returncode, output = run_command(f"git push origin {tag_name}")
        if returncode != 0:
            print_error(f"Failed to push tag: {output}")
            print_warning("Tag created locally but not pushed. Delete it with: git tag -d " + tag_name)
            return False
        print_success(f"Pushed tag to remote: {tag_name}")

    return True


def handle_staging_release(
    force: bool = False, bump_preselect: Optional[str] = None, action_preselect: Optional[str] = None
) -> None:
    """Handle staging release (RC tag creation)."""
    print_header("Staging Release (RC tag)")

    # Check branch
    if not check_branch_validity("staging", force):
        print_warning("Staging release cancelled")
        return

    current_version = get_current_version()
    if not current_version:
        print_error("Could not determine current version")
        sys.exit(1)

    # Ask for version bump
    bump_type = choose_option("Select version bump type:", ["major", "minor", "patch"], preselection=bump_preselect)

    new_version = bump_version(current_version, bump_type)
    print_info(f"Version bump: {current_version} → {new_version}")

    if not confirm_action(
        f"\n{Colors.BOLD}Proceed with {Colors.CYAN}{bump_type}{Colors.RESET} bump to {Colors.GREEN}{new_version}{Colors.RESET}?"
    ):
        print_warning("Staging release cancelled")
        return

    # Update changelog
    repo_root = get_repo_root()
    changelog_path = find_changelog_file(repo_root)
    if changelog_path:
        if not update_changelog(changelog_path, new_version, force):
            print_warning("Staging release cancelled (changelog update failed)")
            return
        if not commit_changelog(changelog_path, new_version):
            print_warning("Failed to commit changelog, but continuing with tag creation")
    else:
        print_warning("No changelog file found; skipping changelog update")

    rc_tag = f"v{new_version}-rc"
    print_info(f"\nRC tag to create: {Colors.BOLD}{Colors.CYAN}{rc_tag}{Colors.RESET}")

    # Ask for push option
    action = choose_option(
        "What would you like to do?",
        ["Create tag locally only", "Create tag and push to remote", "Cancel"],
        preselection=action_preselect,
    )

    if action == "Cancel":
        print_warning("Staging release cancelled")
        return

    push = action == "Create tag and push to remote"

    if create_tag(rc_tag, push=push):
        print_success("\n✓ Staging release prepared!")
        print_info(f"RC tag: {Colors.BOLD}{rc_tag}{Colors.RESET}")
        if not push:
            print_info(f"Push with: {Colors.DIM}git push origin {rc_tag}{Colors.RESET}")
    else:
        print_error("Failed to create staging release")
        sys.exit(1)


def handle_prod_release(
    force: bool = False, action_preselect: Optional[str] = None, rc_tag_preselect: Optional[str] = None
) -> None:
    """Handle production release (promote RC tag to production)."""
    print_header("Production Release (promote RC tag)")

    # Get all RC tags
    returncode, output = run_command("git tag -l 'v*-rc' --sort=-version:refname")

    if returncode != 0 or not output:
        print_error("No RC tags found. Please create a staging release first.")
        sys.exit(1)

    available_rc_tags = output.strip().split("\n")

    if rc_tag_preselect:
        # Validate provided RC tag
        if rc_tag_preselect not in available_rc_tags:
            print_error(f"RC tag '{rc_tag_preselect}' not found.")
            print_info(f"Available RC tags: {', '.join(available_rc_tags[:5])}")
            sys.exit(1)
        rc_tag = rc_tag_preselect
    else:
        # Interactive selection
        rc_tag = choose_option(
            f"Select RC tag to promote (showing {min(10, len(available_rc_tags))} most recent):", available_rc_tags[:10]
        )

    print_info(f"RC tag to promote: {Colors.BOLD}{rc_tag}{Colors.RESET}")

    # Extract version from RC tag
    match = re.match(r"v?(\d+\.\d+\.\d+)-rc", rc_tag)
    if not match:
        print_error(f"Could not parse version from RC tag: {rc_tag}")
        sys.exit(1)

    version = match.group(1)
    prod_tag = f"v{version}"

    print_info(f"Production tag to create: {Colors.BOLD}{Colors.GREEN}{prod_tag}{Colors.RESET}")

    if not confirm_action(f"\nPromote {Colors.CYAN}{rc_tag}{Colors.RESET} to {Colors.GREEN}{prod_tag}{Colors.RESET}?"):
        print_warning("Production release cancelled")
        return

    # Update changelog if present (move version from Unreleased to dated section)
    repo_root = get_repo_root()
    changelog_path = find_changelog_file(repo_root)
    if changelog_path:
        if not update_changelog(changelog_path, version, force):
            print_warning("Production release cancelled (changelog update failed)")
            return
        if not commit_changelog(changelog_path, version):
            print_warning("Failed to commit changelog, but continuing with tag creation")

    # Ask for push option
    action = choose_option(
        "What would you like to do?",
        ["Create tag locally only", "Create tag and push to remote", "Cancel"],
        preselection=action_preselect,
    )

    if action == "Cancel":
        print_warning("Production release cancelled")
        return

    push = action == "Create tag and push to remote"

    if create_tag(prod_tag, push=push):
        print_success("\n✓ Production release prepared!")
        print_info(f"Production tag: {Colors.BOLD}{prod_tag}{Colors.RESET}")
        print_info("RC tag will be promoted by the release pipeline.")
        if not push:
            print_info(f"Push with: {Colors.DIM}git push origin {prod_tag}{Colors.RESET}")
    else:
        print_error("Failed to create production release")
        sys.exit(1)


def main() -> None:
    """Main entry point."""
    import argparse

    parser = argparse.ArgumentParser(
        description="Create release tags for xt-workbench",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog="""
Examples:
  python scripts/release.py           # Interactive menu
  python scripts/release.py --force   # Skip safety checks
        """,
    )
    parser.add_argument(
        "--force",
        action="store_true",
        help="Skip safety checks (branch validation, changelog checks)",
    )
    parser.add_argument(
        "--release-type",
        choices=["staging", "prod"],
        help="Run non-interactively for the given release type",
    )
    parser.add_argument(
        "--bump",
        choices=["major", "minor", "patch"],
        help="Version bump type for staging releases",
    )
    parser.add_argument(
        "--push",
        action="store_true",
        help="Push the created tag to origin (non-interactive mode)",
    )
    parser.add_argument(
        "--yes",
        action="store_true",
        help="Assume yes for all confirmations (non-interactive)",
    )
    parser.add_argument(
        "--rc-tag",
        type=str,
        help="RC tag to promote for production releases (e.g., v1.5.13-rc)",
    )

    args = parser.parse_args()

    global AUTO_CONFIRM
    AUTO_CONFIRM = args.yes

    print_header("XT Workbench Release Manager")

    release_options = ["Staging (RC tag)", "Production (promote RC)", "Cancel"]
    release_preselect = None
    if args.release_type == "staging":
        release_preselect = "Staging (RC tag)"
    elif args.release_type == "prod":
        release_preselect = "Production (promote RC)"

    release_type = choose_option(
        "Select release type:",
        release_options,
        preselection=release_preselect,
    )

    if release_type == "Cancel":
        print_warning("Release cancelled")
        return

    push_action_preselect = "Create tag and push to remote" if args.push else "Create tag locally only"

    if release_type == "Staging (RC tag)":
        bump_preselect = args.bump
        if bump_preselect is None and args.release_type == "staging":
            print_error("--bump is required when --release-type staging is provided")
            sys.exit(1)
        handle_staging_release(force=args.force, bump_preselect=bump_preselect, action_preselect=push_action_preselect)
    elif release_type == "Production (promote RC)":
        if args.rc_tag is None and args.release_type == "prod":
            print_error("--rc-tag is required when --release-type prod is provided")
            sys.exit(1)
        handle_prod_release(force=args.force, action_preselect=push_action_preselect, rc_tag_preselect=args.rc_tag)


if __name__ == "__main__":
    main()



import pathlib

import pytest
import streamlit as st
from pymongo import MongoClient

from xt_st_common.config import StreamlitBaseSettings

# Ensure tests use the test .env
dotenv_path = pathlib.Path(__file__).parent.resolve() / ".env-test"
config = StreamlitBaseSettings(_env_file=dotenv_path)

client = MongoClient(config.MONGO_CONNECTION_STRING)
collection = client.get_database(name=config.DATABASE_NAME)["project"]


@pytest.fixture(autouse=True)
def clear_project_collection():
    """Clear the `project` collection before and after each test to avoid cross-test data leakage."""
    # ensure streamlit caches don't return stale results between tests
    try:
        st.cache_data.clear()
    except Exception:
        pass
    try:
        st.cache_resource.clear()
    except Exception:
        pass

    collection.delete_many({})
    yield
    collection.delete_many({})
